﻿Option Strict On
Option Explicit On

Imports System.IO
Imports System.Runtime.InteropServices


' Basiert auf:
' https://www.autohotkey.com/boards/viewtopic.php?t=80735&start=20
' Posting von iseahound v1/v2

' Headerfiles:
' https://github.com/tpn/winsdk-10/tree/master/Include/10.0.14393.0/winrt

Public Class PdfViewer
    Implements IDisposable

#Region "API"
    <DllImport("Combase.dll", EntryPoint:="WindowsCreateString")>
    <PreserveSig> Private Shared Function WindowsCreateString(<[In], MarshalAs(UnmanagedType.LPWStr)> sourceString As String,
                                                              <[In]> length As Integer,
                                                              <Out> ByRef hString As IntPtr) As Integer
    End Function

    <DllImport("Combase.dll", EntryPoint:="WindowsDeleteString")>
    <PreserveSig> Private Shared Function WindowsDeleteString(<[In]> hString As IntPtr) As Integer
    End Function

    <DllImport("Combase.dll", EntryPoint:="RoGetActivationFactory")>
    <PreserveSig> Private Shared Function RoGetActivationFactory(<[In]> activatableClassId As IntPtr,
                                                                 <[In], MarshalAs(UnmanagedType.LPStruct)> riid As Guid,
                                                                 <Out> ByRef factory As IntPtr) As Integer

    End Function

    <DllImport("Combase.dll", EntryPoint:="RoActivateInstance")>
    <PreserveSig> Private Shared Function RoActivateInstance(<[In]> activatableClassId As IntPtr,
                                                             <Out> ByRef instance As IntPtr) As Integer

    End Function

    <DllImport("Shlwapi.dll", EntryPoint:="SHCreateMemStream")>
    Private Shared Function SHCreateMemStream(<[In]> pInit As IntPtr,
                                              <[In]> cbInit As UInteger) As IntPtr
    End Function


    <DllImport("ShCore.dll", EntryPoint:="CreateRandomAccessStreamOnFile")>
    <PreserveSig> Private Shared Function CreateRandomAccessStreamOnFile(<[In], MarshalAs(UnmanagedType.LPWStr)> filePath As String,
                                                                         <[In]> accessMode As FileAccessMode,
                                                                         <[In], MarshalAs(UnmanagedType.LPStruct)> riid As Guid,
                                                                         <Out> ByRef ppv As IntPtr) As Integer
    End Function

    <DllImport("ShCore.dll", EntryPoint:="CreateRandomAccessStreamOverStream")>
    <PreserveSig> Private Shared Function CreateRandomAccessStreamOverStream(<[In]> pStream As IntPtr,
                                                                             <[In]> options As BSOS_OPTIONS,
                                                                             <[In], MarshalAs(UnmanagedType.LPStruct)> riid As Guid,
                                                                             <Out> ByRef ppv As IntPtr) As Integer
    End Function

    <DllImport("GdiPlus.dll", EntryPoint:="GdipCreateBitmapFromStream")>
    <PreserveSig> Private Shared Function GdipCreateBitmapFromStream(<[In]> pStream As IntPtr,
                                                                     <Out> ByRef pBitmap As IntPtr) As Integer
    End Function

    <DllImport("GdiPlus.dll", EntryPoint:="GdipDisposeImage")>
    <PreserveSig> Private Shared Function GdipDisposeImage(<[In]> pImage As IntPtr) As Integer
    End Function

    <DllImport("GdiPlus.dll", EntryPoint:="GdipCreateHBITMAPFromBitmap")>
    <PreserveSig> Private Shared Function GdipCreateHBITMAPFromBitmap(<[In]> pBitmap As IntPtr,
                                                                      <Out> ByRef hBitmap As IntPtr,
                                                                      <[In]> mBackground As Integer) As Integer
    End Function

    <DllImport("Gdi32.dll", EntryPoint:="DeleteObject")>
    <PreserveSig> Private Shared Function DeleteObject(<[In]> hObject As IntPtr) As Integer
    End Function
#End Region

#Region "Enum"
    Private Enum VTable_Interfaces As Integer

        ' IPdfDocumentStatics
        IPdfDocumentStatics_LoadFromStreamAsync = 8
        IPdfDocumentStatics_LoadFromStreamWithPasswordAsync = 9

        ' IAsyncInfo
        IAsyncInfo_GetStatus = 7
        IAsyncInfo_Close = 10

        ' IAsyncPdfDocument
        IAsyncPdfDocument_GetResults = 8

        ' IPdfDocument
        IPdfDocument_GetPage = 6
        IPdfDocument_GetPageCount = 7
        IPdfDocument_GetIsPasswordProtected = 8

        ' IPdfPage *IClosable
        IPdfPage_RenderWithOptionsToStreamAsync = 7
        IPdfPage_GetIndex = 9
        IPdfPage_GetSize = 10
        IPdfPage_GetDimensions = 11
        IPdfPage_GetRotation = 12
        IPdfPage_GetPreferredZoom = 13

        ' IRandomAccessStream *IClosable
        IRandomAccessStream_GetSize = 6

        ' IPdfPageRenderOptions
        IPdfPageRenderOptions_GetSourceRect = 6
        IPdfPageRenderOptions_PutSourceRect = 7
        IPdfPageRenderOptions_GetDestinationWidth = 8
        IPdfPageRenderOptions_PutDestinationWidth = 9
        IPdfPageRenderOptions_GetDestinationHeight = 10
        IPdfPageRenderOptions_PutDestinationHeight = 11
        IPdfPageRenderOptions_GetBackgroundColor = 12
        IPdfPageRenderOptions_PutBackgroundColor = 13
        IPdfPageRenderOptions_GetIsIgnoringHighContrast = 14
        IPdfPageRenderOptions_PutIsIgnoringHighContrast = 15
        IPdfPageRenderOptions_GetBitmapEncoderId = 16
        IPdfPageRenderOptions_PutBitmapEncoderId = 17

        ' IPdfPageDimensions
        IPdfPageDimensions_GetMediaBox = 6
        IPdfPageDimensions_GetCropBox = 7
        IPdfPageDimensions_GetBleedBox = 8
        IPdfPageDimensions_GetTrimBox = 9
        IPdfPageDimensions_GetArtBox = 10

        ' IClosable
        IClosable_Close = 6
    End Enum

    Private Enum FileAccessMode As Integer
        Read = 0
        ReadWrite = 1
    End Enum

    Private Enum BSOS_OPTIONS As Integer
        BSOS_DEFAULT = 0
        BSOS_PREFERDESTINATIONSTREAM = 1
    End Enum

    Private Enum TrustLevel
        BaseTrust = 0
        PartialTrust = 1
        FullTrust = 2
    End Enum

    Private Enum AsyncStatus As Integer
        Started = 0
        Completed = 1
        Canceled = 2
        [Error] = 3
    End Enum

    Public Enum PdfPageRotation As Integer
        PdfPageRotation_Error = -1
        PdfPageRotation_Normal = 0
        PdfPageRotation_Rotate90 = 1
        PdfPageRotation_Rotate180 = 2
        PdfPageRotation_Rotate270 = 3
    End Enum

    Public Enum SaveFormat As Integer
        Bmp = 0
        Gif = 1
        Png = 2
        Jpg = 3
        Tif = 4
    End Enum
#End Region

#Region "Const"
    Private Const S_OK As Integer = &H0

    Private Const WindowsDataPdfPdfDocument As String = "Windows.Data.Pdf.PdfDocument"
    Private Const WindowsDataPdfPdfPageRenderOptions As String = "Windows.Data.Pdf.PdfPageRenderOptions"

    Private Const IID_IClosable As String = "30d5a829-7fa4-4026-83bb-d75bae4ea99e"
    Private Const IID_IAsyncInfo As String = "00000036-0000-0000-c000-000000000046"
    Private Const IID_IPdfDocumentStatics As String = "433a0b5f-c007-4788-90f2-08143d922599"
    Private Const IID_IRandomAccessStream As String = "905a0fe1-bc53-11df-8c49-001e4fc686da"
    Private Const IID_IPdfPageRenderOptions As String = "3c98056f-b7cf-4c29-9a04-52d90267f425"

    Private Const CLSID_WICPngEncoder As String = "27949969-876a-41d7-9447-568f6a35a4dc"
#End Region

#Region "Class"
    Sub New()
        If GetActivationFactory(WindowsDataPdfPdfDocument,
                                IID_IPdfDocumentStatics,
                                m_pIPdfDocumentStatics) Then
            If m_pIPdfDocumentStatics <> IntPtr.Zero Then
                m_pIStream = SHCreateMemStream(IntPtr.Zero, 0)
                If m_pIStream <> IntPtr.Zero Then
                    If CreateRandomAccessStreamOverStream(m_pIStream, BSOS_OPTIONS.BSOS_DEFAULT,
                                                          New Guid(IID_IRandomAccessStream),
                                                          m_pIRandomAccessStreamOut) = S_OK Then
                        If m_pIRandomAccessStreamOut <> IntPtr.Zero Then m_bolFactoryActivated = True
                    End If
                End If
            End If
        End If
    End Sub
#End Region

#Region "Variablen"
    Private m_strPdfFile As String = String.Empty
    Private m_intPdfPageCount As Integer = 0
    Private m_DisposedValue As Boolean = False
    Private m_bolFactoryActivated As Boolean = False
    Private m_PageInfos As PageInfo() = New PageInfo() {}
    Private m_bolPdfIsPasswordProtected As Boolean = False
    Private m_bolUseSourceRectForAllPdfPages As Boolean = False
    Private m_bolUseScaleFactorForAllPdfPages As Boolean = False
    Private m_bolUseUseBackColorForAllPdfPages As Boolean = False
    Private m_bolUseBitmapEncoderIdForAllPdfPages As Boolean = False
    Private m_bolUseIgnoringHighContrastForAllPdfPages As Boolean = False
    Private m_sngScaleFactorForAllPdfPages As Single = 1
    Private m_intBackColorForAllPdfPages As Integer = &HFFFFFFFF
    Private m_bolIgnoringHighContrastForAllPdfPages As Boolean = True
    Private m_rectSourceRectForAllPdfPages As New RectangleF(0, 0, 0, 0)
    Private m_guidBitmapEncoderIdForAllPdfPages As New Guid(CLSID_WICPngEncoder)
    Private ReadOnly m_pIStream As IntPtr = IntPtr.Zero
    Private ReadOnly m_pIPdfDocumentStatics As IntPtr = IntPtr.Zero
    Private m_pIPdfDocument As IntPtr = IntPtr.Zero
    Private m_pIRandomAccessStreamIn As IntPtr = IntPtr.Zero
    Private m_pIRandomAccessStreamOut As IntPtr = IntPtr.Zero
#End Region

#Region "Structure"
    Private Structure PageInfo
        Dim PageIndex As UInteger
        Dim PageSize As SizeF
        Dim PageRotation As PdfPageRotation
        Dim PagePreferredZoom As Single
        Dim PageMediaBox As RectangleF
        Dim PageCropBox As RectangleF
        Dim PageBleedBox As RectangleF
        Dim PageTrimBox As RectangleF
        Dim PageArtBox As RectangleF
        Dim PageSourceRect As RectangleF
        Dim PageDestinationWidth As Integer
        Dim PageDestinationHeight As Integer
        Dim PageBackgroundColor As Integer
        Dim PageIsIgnoringHighContrast As Boolean
        Dim PageBitmapEncoderId As Guid
    End Structure
#End Region

#Region "Properties"
    Public Property PdfFile As String
        Get
            Return m_strPdfFile
        End Get
        Set(value As String)
            m_strPdfFile = value
        End Set
    End Property

    Public ReadOnly Property LoadPdfFile As Boolean
        Get
            Return LoadPdfFromFile()
        End Get
    End Property

    Public ReadOnly Property SavePdfPage(PageIndex As Integer,
                                         OutputPath As String,
                                         Optional FileName As String = "SavedPdfPage",
                                         Optional SaveAs As SaveFormat = SaveFormat.Png) As Boolean
        Get
            Return SavePdfPageToFile(PageIndex, OutputPath, FileName, SaveAs)
        End Get
    End Property

    Public ReadOnly Property GetIsPdfFactoryActivated As Boolean
        Get
            Return m_bolFactoryActivated
        End Get
    End Property

    Public ReadOnly Property GetPdfPageCount As Integer
        Get
            Return m_intPdfPageCount
        End Get
    End Property

    Public ReadOnly Property GetPdfIsPasswordProtected As Boolean
        Get
            Return m_bolPdfIsPasswordProtected
        End Get
    End Property

    Public ReadOnly Property GetPdfPageAsBitmap(PageIndex As Integer) As Bitmap
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return GetPdfPage(PageIndex)
            Else
                Return Nothing
            End If
        End Get
    End Property

    Public ReadOnly Property GetPdfPageIndex(PageIndex As Integer) As Integer
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return CInt(m_PageInfos(PageIndex).PageIndex)
            Else
                Return -1
            End If
        End Get
    End Property

    Public ReadOnly Property GetPdfPageSize(PageIndex As Integer) As SizeF
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PageSize
            Else
                Return Nothing
            End If
        End Get
    End Property

    Public ReadOnly Property GetPdfPageRotation(PageIndex As Integer) As PdfPageRotation
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PageRotation
            Else
                Return PdfPageRotation.PdfPageRotation_Error
            End If
        End Get
    End Property

    Public ReadOnly Property GetPdfPagePreferredZoom(PageIndex As Integer) As Single
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PagePreferredZoom
            Else
                Return -1
            End If
        End Get
    End Property

    Public ReadOnly Property GetPdfPagMediaBox(PageIndex As Integer) As RectangleF
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PageMediaBox
            Else
                Return Nothing
            End If
        End Get
    End Property

    Public ReadOnly Property GetPdfPagCropBox(PageIndex As Integer) As RectangleF
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PageCropBox
            Else
                Return Nothing
            End If
        End Get
    End Property

    Public ReadOnly Property GetPdfPagBleedBox(PageIndex As Integer) As RectangleF
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PageBleedBox
            Else
                Return Nothing
            End If
        End Get
    End Property

    Public ReadOnly Property GetPdfPagTrimBox(PageIndex As Integer) As RectangleF
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PageTrimBox
            Else
                Return Nothing
            End If
        End Get
    End Property

    Public ReadOnly Property GetPdfPagArtBox(PageIndex As Integer) As RectangleF
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PageArtBox
            Else
                Return Nothing
            End If
        End Get
    End Property

    Public Property PdfPageSourceRect(PageIndex As Integer) As RectangleF
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PageSourceRect
            Else
                Return Nothing
            End If
        End Get
        Set(value As RectangleF)
            m_PageInfos(PageIndex).PageSourceRect = value
        End Set
    End Property

    Public Property PdfPageDestinationWidth(PageIndex As Integer) As Integer
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PageDestinationWidth
            Else
                Return Nothing
            End If
        End Get
        Set(value As Integer)
            m_PageInfos(PageIndex).PageDestinationWidth = value
        End Set
    End Property

    Public Property PdfPageDestinationHeight(PageIndex As Integer) As Integer
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PageDestinationHeight
            Else
                Return Nothing
            End If
        End Get
        Set(value As Integer)
            m_PageInfos(PageIndex).PageDestinationHeight = value
        End Set
    End Property

    Public Property PdfPagePageBackgroundColor(PageIndex As Integer) As Integer
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PageBackgroundColor
            Else
                Return Nothing
            End If
        End Get
        Set(value As Integer)
            m_PageInfos(PageIndex).PageBackgroundColor = value
        End Set
    End Property

    Public Property PdfPageIsIgnoringHighContrast(PageIndex As Integer) As Boolean
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PageIsIgnoringHighContrast
            Else
                Return Nothing
            End If
        End Get
        Set(value As Boolean)
            m_PageInfos(PageIndex).PageIsIgnoringHighContrast = value
        End Set
    End Property

    Public Property PdfPageBitmapEncoderId(PageIndex As Integer) As Guid
        Get
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Return m_PageInfos(PageIndex).PageBitmapEncoderId
            Else
                Return Nothing
            End If
        End Get
        Set(value As Guid)
            m_PageInfos(PageIndex).PageBitmapEncoderId = value
        End Set
    End Property


    Public Property UseUseBackColorForAllPdfPages As Boolean
        Get
            Return m_bolUseUseBackColorForAllPdfPages
        End Get
        Set(value As Boolean)
            m_bolUseUseBackColorForAllPdfPages = value
        End Set
    End Property

    Public Property BackColorForAllPdfPages As Integer
        Get
            Return m_intBackColorForAllPdfPages
        End Get
        Set(value As Integer)
            m_intBackColorForAllPdfPages = value
        End Set
    End Property

    Public Property UseScaleFactorForAllPdfPages As Boolean
        Get
            Return m_bolUseScaleFactorForAllPdfPages
        End Get
        Set(value As Boolean)
            m_bolUseScaleFactorForAllPdfPages = value
        End Set
    End Property

    Public Property ScaleFactorForAllPdfPages As Single
        Get
            Return m_sngScaleFactorForAllPdfPages
        End Get
        Set(value As Single)
            m_sngScaleFactorForAllPdfPages = value
        End Set
    End Property

    Public Property UseIgnoringHighContrastForAllPdfPages As Boolean
        Get
            Return m_bolUseIgnoringHighContrastForAllPdfPages
        End Get
        Set(value As Boolean)
            m_bolUseIgnoringHighContrastForAllPdfPages = value
        End Set
    End Property

    Public Property IgnoringHighContrastForAllPdfPages As Boolean
        Get
            Return m_bolIgnoringHighContrastForAllPdfPages
        End Get
        Set(value As Boolean)
            m_bolIgnoringHighContrastForAllPdfPages = value
        End Set
    End Property

    Public Property UseBitmapEncoderIdForAllPdfPages As Boolean
        Get
            Return m_bolUseBitmapEncoderIdForAllPdfPages
        End Get
        Set(value As Boolean)
            m_bolUseBitmapEncoderIdForAllPdfPages = value
        End Set
    End Property

    Public Property BitmapEncoderIdForAllPdfPages As Guid
        Get
            Return m_guidBitmapEncoderIdForAllPdfPages
        End Get
        Set(value As Guid)
            m_guidBitmapEncoderIdForAllPdfPages = value
        End Set
    End Property

    Public Property UseSourceRectForAllPdfPages As Boolean
        Get
            Return m_bolUseSourceRectForAllPdfPages
        End Get
        Set(value As Boolean)
            m_bolUseSourceRectForAllPdfPages = value
        End Set
    End Property

    Public Property SourceRectForAllPdfPages As RectangleF
        Get
            Return m_rectSourceRectForAllPdfPages
        End Get
        Set(value As RectangleF)
            m_rectSourceRectForAllPdfPages = value
        End Set
    End Property
#End Region

#Region "Private Functions"
    Private Sub CloseAll()
        Release(m_pIPdfDocument)
        CloseAndRelease(m_pIRandomAccessStreamIn)



    End Sub

    Private Function LoadPdfFromFile() As Boolean
        Dim bolRet As Boolean = False
        If m_bolFactoryActivated Then
            CloseAll()
            If m_pIPdfDocumentStatics <> IntPtr.Zero Then
                If CreateRandomAccessStreamOnFile(m_strPdfFile, FileAccessMode.Read,
                                                  New Guid(IID_IRandomAccessStream),
                                                  m_pIRandomAccessStreamIn) = S_OK Then
                    If m_pIRandomAccessStreamIn <> IntPtr.Zero Then
                        Dim pIAsyncPdfDocument As IntPtr = IntPtr.Zero
                        If PdfDocumentStatics_LoadFromStreamAsync(m_pIPdfDocumentStatics,
                                                                  m_pIRandomAccessStreamIn,
                                                                  pIAsyncPdfDocument) Then
                            If pIAsyncPdfDocument <> IntPtr.Zero Then
                                If Not WaitForAsync(pIAsyncPdfDocument) Then
                                    Release(pIAsyncPdfDocument)
                                    Dim strPassword As String = String.Empty
                                    If InputQuery("PdfViewer", "'" & Path.GetFileName(PdfFile) & "' is protected. " &
                                                  "Please enter a Document Open Password.", strPassword) = DialogResult.OK Then
                                        If Not String.IsNullOrEmpty(strPassword) Then
                                            Dim hString As IntPtr = IntPtr.Zero
                                            If WindowsCreateString(strPassword, strPassword.Length, hString) = S_OK Then
                                                If PdfDocumentStatics_LoadFromStreamWithPasswordAsync(m_pIPdfDocumentStatics,
                                                                                                      m_pIRandomAccessStreamIn,
                                                                                                      hString, pIAsyncPdfDocument) Then
                                                    If pIAsyncPdfDocument <> IntPtr.Zero Then
                                                        If Not WaitForAsync(pIAsyncPdfDocument) Then
                                                            Release(pIAsyncPdfDocument)
                                                            MessageBox.Show("Wrong Password.", "PdfViewer",
                                                                            MessageBoxButtons.OK, MessageBoxIcon.Error)
                                                        End If
                                                    End If
                                                End If
                                                WindowsDeleteString(hString)
                                            End If
                                        Else
                                            MessageBox.Show("Password cannot be null or empty.", "PdfViewer",
                                                            MessageBoxButtons.OK, MessageBoxIcon.Error)
                                        End If
                                    End If
                                End If
                            End If
                        End If
                        If pIAsyncPdfDocument <> IntPtr.Zero Then
                            If AsyncPdfDocument_GetResults(pIAsyncPdfDocument, m_pIPdfDocument) Then
                                If m_pIPdfDocument <> IntPtr.Zero Then
                                    PdfDocument_GetPageCount(m_pIPdfDocument, m_intPdfPageCount)
                                    PdfDocument_GetIsPasswordProtected(m_pIPdfDocument, m_bolPdfIsPasswordProtected)
                                    GetPdfPageInfos()
                                    bolRet = True
                                End If
                            End If
                            Release(pIAsyncPdfDocument)
                        Else
                            CloseAll()
                        End If
                    End If
                End If
            End If
        End If
        Return bolRet
    End Function

    Private Function SavePdfPageToFile(PageIndex As Integer,
                                       OutputPath As String,
                                       FileName As String,
                                       SaveAs As SaveFormat) As Boolean
        Dim bolRet As Boolean
        If m_bolFactoryActivated Then
            If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
                Dim bmpPdfPage As Bitmap = Me.GetPdfPageAsBitmap(PageIndex)
                If bmpPdfPage IsNot Nothing Then
                    Dim imgFormat As Imaging.ImageFormat = Nothing
                    Dim strFileExt As String = String.Empty
                    Select Case SaveAs
                        Case SaveFormat.Bmp
                            strFileExt = ".bmp"
                            imgFormat = Imaging.ImageFormat.Bmp
                        Case SaveFormat.Gif
                            strFileExt = ".gif"
                            imgFormat = Imaging.ImageFormat.Gif
                        Case SaveFormat.Png
                            strFileExt = ".png"
                            imgFormat = Imaging.ImageFormat.Png
                        Case SaveFormat.Jpg
                            strFileExt = ".jpg"
                            imgFormat = Imaging.ImageFormat.Jpeg
                        Case SaveFormat.Tif
                            strFileExt = ".tif"
                            imgFormat = Imaging.ImageFormat.Tiff
                    End Select
                    Dim strPath = Path.Combine(OutputPath, FileName) & "_" &
                                  (PageIndex + 1).ToString.PadLeft(Me.GetPdfPageCount.ToString.Length, "0"c) &
                                  strFileExt
                    bmpPdfPage.Save(strPath, imgFormat)
                    bmpPdfPage.Dispose()
                    If File.Exists(strPath) Then bolRet = True
                End If
            End If
        End If
        Return bolRet
    End Function

    Private Sub GetPdfPageInfos()
        Dim intPageCount = Me.GetPdfPageCount
        If intPageCount > 0 Then
            If m_pIPdfDocument <> IntPtr.Zero Then
                Array.Resize(m_PageInfos, intPageCount)
                For intPageIndex As Integer = 0 To intPageCount - 1
                    Dim pIPdfPage As IntPtr = IntPtr.Zero
                    If PdfDocument_GetPage(m_pIPdfDocument, intPageIndex, pIPdfPage) Then
                        If pIPdfPage <> IntPtr.Zero Then
                            PdfPage_GetIndex(pIPdfPage, m_PageInfos(intPageIndex).PageIndex)
                            PdfPage_GetSize(pIPdfPage, m_PageInfos(intPageIndex).PageSize)
                            Dim pIPdfPageDimensions As IntPtr = IntPtr.Zero
                            If PdfPage_GetDimensions(pIPdfPage, pIPdfPageDimensions) Then
                                If pIPdfPageDimensions <> IntPtr.Zero Then
                                    PdfPageDimensions_GetMediaBox(pIPdfPageDimensions, m_PageInfos(intPageIndex).PageMediaBox)
                                    PdfPageDimensions_GetCropBox(pIPdfPageDimensions, m_PageInfos(intPageIndex).PageCropBox)
                                    PdfPageDimensions_GetBleedBox(pIPdfPageDimensions, m_PageInfos(intPageIndex).PageBleedBox)
                                    PdfPageDimensions_GetTrimBox(pIPdfPageDimensions, m_PageInfos(intPageIndex).PageTrimBox)
                                    PdfPageDimensions_GetArtBox(pIPdfPageDimensions, m_PageInfos(intPageIndex).PageArtBox)
                                    Release(pIPdfPageDimensions)
                                End If
                            End If
                            PdfPage_GetRotation(pIPdfPage, m_PageInfos(intPageIndex).PageRotation)
                            PdfPage_GetPreferredZoom(pIPdfPage, m_PageInfos(intPageIndex).PagePreferredZoom)
                            Dim pIPdfPageRenderOptions As IntPtr = IntPtr.Zero
                            If GetActivateInstance(WindowsDataPdfPdfPageRenderOptions,
                                                   IID_IPdfPageRenderOptions,
                                                   pIPdfPageRenderOptions) Then
                                If pIPdfPageRenderOptions <> IntPtr.Zero Then
                                    PdfPageRenderOptions_GetSourceRect(pIPdfPageRenderOptions, m_PageInfos(intPageIndex).PageSourceRect)
                                    PdfPageRenderOptions_GetDestinationWidth(pIPdfPageRenderOptions, m_PageInfos(intPageIndex).PageDestinationWidth)
                                    PdfPageRenderOptions_GetDestinationHeight(pIPdfPageRenderOptions, m_PageInfos(intPageIndex).PageDestinationHeight)
                                    PdfPageRenderOptions_GetBackgroundColor(pIPdfPageRenderOptions, m_PageInfos(intPageIndex).PageBackgroundColor)
                                    PdfPageRenderOptions_GetIsIgnoringHighContrast(pIPdfPageRenderOptions, m_PageInfos(intPageIndex).PageIsIgnoringHighContrast)
                                    PdfPageRenderOptions_GetBitmapEncoderId(pIPdfPageRenderOptions, m_PageInfos(intPageIndex).PageBitmapEncoderId)
                                    Release(pIPdfPageRenderOptions)
                                End If
                            End If
                            CloseAndRelease(pIPdfPage)
                        End If
                    End If
                Next
            End If
        End If
    End Sub

    Private Function GetPdfPage(PageIndex As Integer) As Bitmap
        Dim bmpRet As Bitmap = Nothing
        If PageIndex >= 0 And PageIndex < Me.GetPdfPageCount Then
            If m_pIPdfDocument <> IntPtr.Zero Then
                Dim pIPdfPage As IntPtr = IntPtr.Zero
                If PdfDocument_GetPage(m_pIPdfDocument, PageIndex, pIPdfPage) Then
                    If pIPdfPage <> IntPtr.Zero Then
                        If m_pIStream <> IntPtr.Zero Then
                            If m_pIRandomAccessStreamOut <> IntPtr.Zero Then
                                Dim pIPdfPageRenderOptions As IntPtr = IntPtr.Zero
                                If GetActivateInstance(WindowsDataPdfPdfPageRenderOptions,
                                                       IID_IPdfPageRenderOptions,
                                                       pIPdfPageRenderOptions) Then
                                    If pIPdfPageRenderOptions <> IntPtr.Zero Then
                                        If m_bolUseScaleFactorForAllPdfPages Then

                                            PdfPageRenderOptions_PutDestinationWidth(pIPdfPageRenderOptions,
                                                                                     CInt(m_PageInfos(PageIndex).PageSize.Width *
                                                                                          m_sngScaleFactorForAllPdfPages))

                                            PdfPageRenderOptions_PutDestinationHeight(pIPdfPageRenderOptions,
                                                                                      CInt(m_PageInfos(PageIndex).PageSize.Height *
                                                                                           m_sngScaleFactorForAllPdfPages))

                                        Else
                                            PdfPageRenderOptions_PutDestinationWidth(pIPdfPageRenderOptions,
                                                                                     m_PageInfos(PageIndex).PageDestinationWidth)
                                            PdfPageRenderOptions_PutDestinationHeight(pIPdfPageRenderOptions,
                                                                                      m_PageInfos(PageIndex).PageDestinationHeight)
                                        End If
                                        If m_bolUseUseBackColorForAllPdfPages Then
                                            PdfPageRenderOptions_PutBackgroundColor(pIPdfPageRenderOptions,
                                                                                    m_intBackColorForAllPdfPages)
                                        Else
                                            PdfPageRenderOptions_PutBackgroundColor(pIPdfPageRenderOptions,
                                                                                    m_PageInfos(PageIndex).PageBackgroundColor)
                                        End If
                                        If m_bolUseIgnoringHighContrastForAllPdfPages Then
                                            PdfPageRenderOptions_PutIsIgnoringHighContrast(pIPdfPageRenderOptions,
                                                                                           m_bolIgnoringHighContrastForAllPdfPages)
                                        Else
                                            PdfPageRenderOptions_PutIsIgnoringHighContrast(pIPdfPageRenderOptions,
                                                                                           m_PageInfos(PageIndex).PageIsIgnoringHighContrast)
                                        End If
                                        If m_bolUseBitmapEncoderIdForAllPdfPages Then
                                            PdfPageRenderOptions_PutBitmapEncoderId(pIPdfPageRenderOptions,
                                                                                    m_guidBitmapEncoderIdForAllPdfPages)
                                        Else
                                            PdfPageRenderOptions_PutBitmapEncoderId(pIPdfPageRenderOptions,
                                                                                    m_PageInfos(PageIndex).PageBitmapEncoderId)
                                        End If
                                        If m_bolUseSourceRectForAllPdfPages Then
                                            PdfPageRenderOptions_PutSourceRect(pIPdfPageRenderOptions,
                                                                               m_rectSourceRectForAllPdfPages)
                                        Else
                                            PdfPageRenderOptions_PutSourceRect(pIPdfPageRenderOptions,
                                                                               m_PageInfos(PageIndex).PageSourceRect)
                                        End If
                                        Dim pIAsyncAction As IntPtr = IntPtr.Zero
                                        If PdfPage_RenderWithOptionsToStreamAsync(pIPdfPage, m_pIRandomAccessStreamOut,
                                                                                  pIPdfPageRenderOptions, pIAsyncAction) Then
                                            If pIAsyncAction <> IntPtr.Zero Then
                                                If WaitForAsync(pIAsyncAction) Then
                                                    Dim ulSize As ULong
                                                    If RandomAccessStream_GetSize(m_pIRandomAccessStreamOut, ulSize) Then
                                                        If ulSize > 0 Then
                                                            bmpRet = GetBitmapFromStream(m_pIStream)
                                                        End If
                                                    End If
                                                End If
                                                Release(pIAsyncAction)
                                            End If
                                        End If
                                        Release(pIPdfPageRenderOptions)
                                    End If
                                End If
                            End If
                        End If
                        CloseAndRelease(pIPdfPage)
                    End If
                End If
            End If
        End If
        Return bmpRet
    End Function

    Private Function GetActivationFactory(ClassName As String,
                                          IID As String,
                                          ByRef pFactory As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        Dim hString As IntPtr = IntPtr.Zero
        If WindowsCreateString(ClassName, ClassName.Length, hString) = S_OK Then
            If hString <> IntPtr.Zero Then
                If RoGetActivationFactory(hString, New Guid(IID), pFactory) = S_OK Then bolRet = True
                WindowsDeleteString(hString)
            End If
        End If
        Return bolRet
    End Function

    Private Function GetActivateInstance(ClassName As String,
                                         IID As String,
                                         ByRef pInstance As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        Dim hString As IntPtr = IntPtr.Zero
        If WindowsCreateString(ClassName, ClassName.Length, hString) = S_OK Then
            If hString <> IntPtr.Zero Then
                Dim pIInspectable As IntPtr = IntPtr.Zero
                If RoActivateInstance(hString, pIInspectable) = S_OK Then
                    If pIInspectable <> IntPtr.Zero Then
                        If QueryInterface(pIInspectable, IID, pInstance) Then
                            If pInstance <> IntPtr.Zero Then bolRet = True
                        End If
                        Release(pIInspectable)
                    End If
                End If
                WindowsDeleteString(hString)
            End If
        End If
        Return bolRet
    End Function

    Private Function GetBitmapFromStream(pStream As IntPtr) As Bitmap
        Dim bmpRet As Bitmap = Nothing
        Dim pBitmap As IntPtr = IntPtr.Zero
        If GdipCreateBitmapFromStream(pStream, pBitmap) = S_OK Then
            Dim hBitmap As IntPtr = IntPtr.Zero
            If GdipCreateHBITMAPFromBitmap(pBitmap, hBitmap, 0) = S_OK Then
                bmpRet = Bitmap.FromHbitmap(hBitmap)
                DeleteObject(hBitmap)
            End If
            GdipDisposeImage(pBitmap)
        End If
        Return bmpRet
    End Function

    Private Function WaitForAsync(pInterface As IntPtr) As Boolean
        Dim bolret As Boolean = False
        If pInterface <> IntPtr.Zero Then
            Dim pIAsyncInfo As IntPtr = IntPtr.Zero
            If QueryInterface(pInterface, IID_IAsyncInfo, pIAsyncInfo) Then
                If pIAsyncInfo <> IntPtr.Zero Then
                    Dim eStatus As AsyncStatus
                    Do
                        AsyncInfo_GetStatus(pIAsyncInfo, eStatus)
                    Loop While eStatus = AsyncStatus.Started
                    If eStatus = AsyncStatus.Completed Then
                        bolret = True
                    Else
                        AsyncInfo_Close(pIAsyncInfo)
                    End If
                    Release(pIAsyncInfo)
                End If
            End If
        End If
        Return bolret
    End Function

    Private Function InputQuery(Caption As String,
                                Prompt As String,
                                ByRef Value As String) As DialogResult
        Dim Frm As New Form With {
            .AutoScaleMode = AutoScaleMode.Font,
            .Font = SystemFonts.IconTitleFont,
            .FormBorderStyle = FormBorderStyle.FixedDialog,
            .MinimizeBox = False,
            .MaximizeBox = False,
            .Text = Caption,
            .StartPosition = FormStartPosition.CenterParent,
            .ClientSize = New Size(430, 100)
        }
        Dim LblPrompt = New Label With {
            .Parent = Frm,
            .AutoSize = True,
            .Left = 8,
            .Top = 8,
            .Text = Prompt
        }
        Dim EdInput = New TextBox With {
            .Parent = Frm,
            .Left = LblPrompt.Left,
            .Top = LblPrompt.Top + LblPrompt.Height + 8,
            .Width = Frm.ClientSize.Width - 16,
            .Text = Value,
            .PasswordChar = "*"c
        }
        EdInput.SelectAll()
        Dim ButtonTop As Integer = EdInput.Top + EdInput.Height + 8
        Dim ButtonSize As New Size(100, 25)
        Dim BtnOk = New Button With {
            .Parent = Frm,
            .Text = "&OK",
            .DialogResult = DialogResult.OK,
            .Location = New Point((Frm.ClientSize.Width \ 2) - (ButtonSize.Width + 2), ButtonTop),
            .Size = ButtonSize
        }
        Dim BtnCancel = New Button With {
            .Parent = Frm,
            .Text = "&Cancel",
            .DialogResult = DialogResult.Cancel,
            .Location = New Point((Frm.ClientSize.Width \ 2) + 2, ButtonTop),
            .Size = ButtonSize
        }
        Frm.AcceptButton = BtnOk
        Frm.CancelButton = BtnCancel

        If Frm.ShowDialog() = DialogResult.OK Then
            Value = EdInput.Text
            Return DialogResult.OK
        Else
            Return DialogResult.Cancel
        End If
    End Function

    Private Function GetVTablePtr(pInterface As IntPtr,
                                  intMethodNumber As VTable_Interfaces,
                                  ByRef pVTablePtr As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        If pInterface <> IntPtr.Zero AndAlso intMethodNumber >= 0 Then
            Dim pVTablePtrs As IntPtr() = New IntPtr(0) {}
            Marshal.Copy(IntPtr.Add(Marshal.ReadIntPtr(pInterface),
                                    IntPtr.Size * intMethodNumber),
                                    pVTablePtrs, 0, 1)
            pVTablePtr = pVTablePtrs(0)
            bolRet = True
        End If
        Return bolRet
    End Function

    Private Sub Release(ByRef pInterface As IntPtr)
        If pInterface <> IntPtr.Zero Then
            Marshal.Release(pInterface)
            pInterface = IntPtr.Zero
        End If
    End Sub

    Private Sub CloseAndRelease(ByRef pInterface As IntPtr)
        If pInterface <> IntPtr.Zero Then
            Dim pIClosable As IntPtr = IntPtr.Zero
            If QueryInterface(pInterface, IID_IClosable, pIClosable) Then
                If pIClosable <> IntPtr.Zero Then
                    If Closable_Close(pIClosable) Then
                        Marshal.Release(pIClosable)
                        pInterface = IntPtr.Zero
                    End If
                End If
            End If
        End If
    End Sub

    Private Function QueryInterface(pInterface As IntPtr,
                                    IID As String,
                                    ByRef pIOut As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        If pInterface <> IntPtr.Zero Then
            If Marshal.QueryInterface(pInterface, New Guid(IID), pIOut) = S_OK Then
                If pIOut <> IntPtr.Zero Then bolRet = True
            End If
        End If
        Return bolRet
    End Function
#End Region

#Region "Interface Delegates"
    Private Delegate Function IPdfDocumentStatics_LoadFromStreamAsync(<[In]> this As IntPtr,
                                                                      <[In]> inputStream As IntPtr,
                                                                      <Out> ByRef asyncInfo As IntPtr) As Integer

    Private Delegate Function IPdfDocumentStatics_LoadFromStreamWithPasswordAsync(<[In]> this As IntPtr,
                                                                                  <[In]> inputStream As IntPtr,
                                                                                  <[In]> password As IntPtr,
                                                                                  <Out> ByRef asyncInfo As IntPtr) As Integer

    Private Delegate Function IAsyncInfo_GetStatus(<[In]> this As IntPtr,
                                                    <Out> ByRef status As AsyncStatus) As Integer

    Private Delegate Function IAsyncInfo_Close(<[In]> this As IntPtr) As Integer


    Private Delegate Function IAsyncPdfDocument_GetResults(<[In]> this As IntPtr,
                                                           <Out> ByRef results As IntPtr) As Integer

    Private Delegate Function IPdfDocument_GetPage(<[In]> this As IntPtr,
                                                   <[In]> pageIndex As Integer,
                                                   <Out> ByRef pdfPage As IntPtr) As Integer

    Private Delegate Function IPdfDocument_GetPageCount(<[In]> this As IntPtr,
                                                         <Out> ByRef value As Integer) As Integer

    Private Delegate Function IPdfDocument_GetIsPasswordProtected(<[In]> this As IntPtr,
                                                                   <Out, MarshalAs(UnmanagedType.Bool)> ByRef value As Boolean) As Integer

    Private Delegate Function IPdfPage_RenderWithOptionsToStreamAsync(<[In]> this As IntPtr,
                                                                      <[In]> outputStream As IntPtr,
                                                                      <[In]> options As IntPtr,
                                                                      <Out> ByRef asyncInfo As IntPtr) As Integer

    Private Delegate Function IPdfPage_GetIndex(<[In]> this As IntPtr,
                                                 <Out> ByRef value As UInteger) As Integer

    Private Delegate Function IPdfPage_GetSize(<[In]> this As IntPtr,
                                               <Out> ByRef value As SizeF) As Integer

    Private Delegate Function IPdfPage_GetDimensions(<[In]> this As IntPtr,
                                                     <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function IPdfPage_GetRotation(<[In]> this As IntPtr,
                                                   <Out> ByRef value As PdfPageRotation) As Integer

    Private Delegate Function IPdfPage_GetPreferredZoom(<[In]> this As IntPtr,
                                                        <Out> ByRef value As Single) As Integer

    Private Delegate Function IRandomAccessStream_GetSize(<[In]> this As IntPtr,
                                                          <Out> ByRef value As ULong) As Integer

    Private Delegate Function IPdfPageRenderOptions_GetSourceRect(<[In]> this As IntPtr,
                                                                  <Out> ByRef value As RectangleF) As Integer

    Private Delegate Function IPdfPageRenderOptions_PutSourceRect(<[In]> this As IntPtr,
                                                                  <[In]> value As RectangleF) As Integer

    Private Delegate Function IPdfPageRenderOptions_GetDestinationWidth(<[In]> this As IntPtr,
                                                                        <Out> ByRef value As Integer) As Integer

    Private Delegate Function IPdfPageRenderOptions_PutDestinationWidth(<[In]> this As IntPtr,
                                                                        <[In]> value As Integer) As Integer

    Private Delegate Function IPdfPageRenderOptions_GetDestinationHeight(<[In]> this As IntPtr,
                                                                         <Out> ByRef value As Integer) As Integer

    Private Delegate Function IPdfPageRenderOptions_PutDestinationHeight(<[In]> this As IntPtr,
                                                                         <[In]> value As Integer) As Integer

    Private Delegate Function IPdfPageRenderOptions_GetBackgroundColor(<[In]> this As IntPtr,
                                                                       <Out> ByRef value As Integer) As Integer

    Private Delegate Function IPdfPageRenderOptions_PutBackgroundColor(<[In]> this As IntPtr,
                                                                       <[In]> value As Integer) As Integer

    Private Delegate Function IPdfPageRenderOptions_GetIsIgnoringHighContrast(<[In]> this As IntPtr,
                                                                              <Out, MarshalAs(UnmanagedType.Bool)> ByRef value As Boolean) As Integer

    Private Delegate Function IPdfPageRenderOptions_PutIsIgnoringHighContrast(<[In]> this As IntPtr,
                                                                              <[In], MarshalAs(UnmanagedType.Bool)> value As Boolean) As Integer

    Private Delegate Function IPdfPageRenderOptions_GetBitmapEncoderId(<[In]> this As IntPtr,
                                                                       <Out> ByRef value As Guid) As Integer

    Private Delegate Function IPdfPageRenderOptions_PutBitmapEncoderId(<[In]> this As IntPtr,
                                                                       <[In]> value As Guid) As Integer

    Private Delegate Function IPdfPageDimensions_GetMediaBox(<[In]> this As IntPtr,
                                                             <Out> ByRef value As RectangleF) As Integer

    Private Delegate Function IPdfPageDimensions_GetCropBox(<[In]> this As IntPtr,
                                                            <Out> ByRef value As RectangleF) As Integer

    Private Delegate Function IPdfPageDimensions_GetBleedBox(<[In]> this As IntPtr,
                                                             <Out> ByRef value As RectangleF) As Integer

    Private Delegate Function IPdfPageDimensions_GetTrimBox(<[In]> this As IntPtr,
                                                            <Out> ByRef value As RectangleF) As Integer

    Private Delegate Function IPdfPageDimensions_GetArtBox(<[In]> this As IntPtr,
                                                           <Out> ByRef value As RectangleF) As Integer

    Private Delegate Function IClosable_Close(<[In]> this As IntPtr) As Integer
#End Region

#Region "Interface Functions"
    Private Function PdfDocumentStatics_LoadFromStreamAsync(this As IntPtr,
                                                            inputStream As IntPtr,
                                                            ByRef asyncInfo As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfDocumentStatics_LoadFromStreamAsync, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfDocumentStatics_LoadFromStreamAsync)),
                                                           IPdfDocumentStatics_LoadFromStreamAsync).Invoke(this, inputStream, asyncInfo) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfDocumentStatics_LoadFromStreamWithPasswordAsync(this As IntPtr,
                                                                        inputStream As IntPtr,
                                                                        password As IntPtr,
                                                                        ByRef asyncInfo As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfDocumentStatics_LoadFromStreamWithPasswordAsync, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfDocumentStatics_LoadFromStreamWithPasswordAsync)),
                                                           IPdfDocumentStatics_LoadFromStreamWithPasswordAsync).Invoke(this, inputStream, password, asyncInfo) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function AsyncInfo_GetStatus(this As IntPtr,
                                           ByRef status As AsyncStatus) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IAsyncInfo_GetStatus, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IAsyncInfo_GetStatus)),
                                                           IAsyncInfo_GetStatus).Invoke(this, status) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function AsyncInfo_Close(this As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IAsyncInfo_Close, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IAsyncInfo_Close)),
                                                           IAsyncInfo_Close).Invoke(this) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function AsyncPdfDocument_GetResults(this As IntPtr,
                                                  ByRef results As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IAsyncPdfDocument_GetResults, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IAsyncPdfDocument_GetResults)),
                                                           IAsyncPdfDocument_GetResults).Invoke(this, results) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfDocument_GetPage(this As IntPtr,
                                         pageIndex As Integer,
                                         ByRef pdfPage As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfDocument_GetPage, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfDocument_GetPage)),
                                                           IPdfDocument_GetPage).Invoke(this, pageIndex, pdfPage) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfDocument_GetPageCount(this As IntPtr,
                                              ByRef value As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfDocument_GetPageCount, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfDocument_GetPageCount)),
                                                           IPdfDocument_GetPageCount).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfDocument_GetIsPasswordProtected(this As IntPtr,
                                                        ByRef value As Boolean) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfDocument_GetIsPasswordProtected, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfDocument_GetIsPasswordProtected)),
                                                           IPdfDocument_GetIsPasswordProtected).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPage_RenderWithOptionsToStreamAsync(this As IntPtr,
                                                            outputStream As IntPtr,
                                                            options As IntPtr,
                                                            ByRef asyncInfo As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPage_RenderWithOptionsToStreamAsync, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPage_RenderWithOptionsToStreamAsync)),
                                                           IPdfPage_RenderWithOptionsToStreamAsync).Invoke(this, outputStream, options, asyncInfo) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPage_GetIndex(this As IntPtr,
                                      ByRef value As UInteger) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPage_GetIndex, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPage_GetIndex)),
                                                           IPdfPage_GetIndex).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPage_GetSize(this As IntPtr,
                                     ByRef value As SizeF) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPage_GetSize, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPage_GetSize)),
                                                           IPdfPage_GetSize).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPage_GetDimensions(this As IntPtr,
                                           ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPage_GetDimensions, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPage_GetDimensions)),
                                                           IPdfPage_GetDimensions).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPage_GetRotation(this As IntPtr,
                                         ByRef value As PdfPageRotation) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPage_GetRotation, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPage_GetRotation)),
                                                           IPdfPage_GetRotation).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPage_GetPreferredZoom(this As IntPtr,
                                              ByRef value As Single) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPage_GetPreferredZoom, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPage_GetPreferredZoom)),
                                                           IPdfPage_GetPreferredZoom).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function RandomAccessStream_GetSize(this As IntPtr,
                                                ByRef value As ULong) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IRandomAccessStream_GetSize, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IRandomAccessStream_GetSize)),
                                                           IRandomAccessStream_GetSize).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageRenderOptions_GetSourceRect(this As IntPtr,
                                                        ByRef value As RectangleF) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageRenderOptions_GetSourceRect, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageRenderOptions_GetSourceRect)),
                                                           IPdfPageRenderOptions_GetSourceRect).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageRenderOptions_PutSourceRect(this As IntPtr,
                                                        value As RectangleF) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageRenderOptions_PutSourceRect, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageRenderOptions_PutSourceRect)),
                                                           IPdfPageRenderOptions_PutSourceRect).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageRenderOptions_GetDestinationWidth(this As IntPtr,
                                                              ByRef value As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageRenderOptions_GetDestinationWidth, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageRenderOptions_GetDestinationWidth)),
                                                           IPdfPageRenderOptions_GetDestinationWidth).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageRenderOptions_PutDestinationWidth(this As IntPtr,
                                                              value As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageRenderOptions_PutDestinationWidth, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageRenderOptions_PutDestinationWidth)),
                                                           IPdfPageRenderOptions_PutDestinationWidth).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageRenderOptions_GetDestinationHeight(this As IntPtr,
                                                               ByRef value As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageRenderOptions_GetDestinationHeight, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageRenderOptions_GetDestinationHeight)),
                                                           IPdfPageRenderOptions_GetDestinationHeight).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageRenderOptions_PutDestinationHeight(this As IntPtr,
                                                               value As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageRenderOptions_PutDestinationHeight, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageRenderOptions_PutDestinationHeight)),
                                                           IPdfPageRenderOptions_PutDestinationHeight).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageRenderOptions_GetBackgroundColor(this As IntPtr,
                                                             ByRef value As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageRenderOptions_GetBackgroundColor, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageRenderOptions_GetBackgroundColor)),
                                                           IPdfPageRenderOptions_GetBackgroundColor).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageRenderOptions_PutBackgroundColor(this As IntPtr,
                                                             value As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageRenderOptions_PutBackgroundColor, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageRenderOptions_PutBackgroundColor)),
                                                           IPdfPageRenderOptions_PutBackgroundColor).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageRenderOptions_GetIsIgnoringHighContrast(this As IntPtr,
                                                                    ByRef value As Boolean) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageRenderOptions_GetIsIgnoringHighContrast, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageRenderOptions_GetIsIgnoringHighContrast)),
                                                           IPdfPageRenderOptions_GetIsIgnoringHighContrast).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageRenderOptions_PutIsIgnoringHighContrast(this As IntPtr,
                                                                    value As Boolean) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageRenderOptions_PutIsIgnoringHighContrast, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageRenderOptions_PutIsIgnoringHighContrast)),
                                                           IPdfPageRenderOptions_PutIsIgnoringHighContrast).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageRenderOptions_GetBitmapEncoderId(this As IntPtr,
                                                             ByRef value As Guid) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageRenderOptions_GetBitmapEncoderId, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageRenderOptions_GetBitmapEncoderId)),
                                                           IPdfPageRenderOptions_GetBitmapEncoderId).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageRenderOptions_PutBitmapEncoderId(this As IntPtr,
                                                             value As Guid) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageRenderOptions_PutBitmapEncoderId, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageRenderOptions_PutBitmapEncoderId)),
                                                           IPdfPageRenderOptions_PutBitmapEncoderId).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageDimensions_GetMediaBox(this As IntPtr,
                                                   ByRef value As RectangleF) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageDimensions_GetMediaBox, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageDimensions_GetMediaBox)),
                                                           IPdfPageDimensions_GetMediaBox).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageDimensions_GetCropBox(this As IntPtr,
                                                  ByRef value As RectangleF) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageDimensions_GetCropBox, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageDimensions_GetCropBox)),
                                                           IPdfPageDimensions_GetCropBox).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageDimensions_GetBleedBox(this As IntPtr,
                                                   ByRef value As RectangleF) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageDimensions_GetBleedBox, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageDimensions_GetBleedBox)),
                                                           IPdfPageDimensions_GetBleedBox).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageDimensions_GetTrimBox(this As IntPtr,
                                                  ByRef value As RectangleF) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageDimensions_GetTrimBox, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageDimensions_GetTrimBox)),
                                                           IPdfPageDimensions_GetTrimBox).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function PdfPageDimensions_GetArtBox(this As IntPtr,
                                                 ByRef value As RectangleF) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IPdfPageDimensions_GetArtBox, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IPdfPageDimensions_GetArtBox)),
                                                           IPdfPageDimensions_GetArtBox).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function Closable_Close(this As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IClosable_Close, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IClosable_Close)),
                                                           IClosable_Close).Invoke(this) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function
#End Region

#Region "Implements IDisposable"
    Protected Overridable Sub Dispose(disposing As Boolean)
        If Not m_DisposedValue Then
            If disposing Then
            End If

            CloseAll()
            CloseAndRelease(m_pIRandomAccessStreamOut)
            Release(m_pIStream)
            Release(m_pIPdfDocumentStatics)
            m_bolFactoryActivated = False
            m_DisposedValue = True
        End If
    End Sub

    Public Sub Dispose() Implements IDisposable.Dispose
        Dispose(disposing:=True)
        GC.SuppressFinalize(Me)
    End Sub
#End Region

End Class
